use std::{
    fmt::Display,
    io::{Write, stderr},
    sync::OnceLock,
};

use anstream::AutoStream;
use anstyle::{AnsiColor, Effects, Style};
use clap::ValueEnum;
use crossterm::{ExecutableCommand, cursor::MoveToPreviousLine};

pub const ERROR: Style = AnsiColor::Red.on_default().effects(Effects::BOLD);
pub const GOOD: Style = AnsiColor::Green.on_default().effects(Effects::BOLD);

#[derive(Copy, Clone, Debug, PartialEq, Eq, ValueEnum)]
pub enum Color {
    Auto,
    Always,
    Never,
}

impl From<Color> for anstream::ColorChoice {
    fn from(value: Color) -> Self {
        match value {
            Color::Always => Self::Always,
            Color::Never => Self::Never,
            Color::Auto => Self::Auto,
        }
    }
}

#[derive(Debug)]
pub struct Console(anstream::ColorChoice);

impl Console {
    fn new(color: Color) -> Self {
        Self(color.into())
    }

    pub fn print(&self, status: &str, style: &Style, msg: impl Display) {
        let mut stream = AutoStream::new(stderr(), self.0);
        let _ = writeln!(stream, "{style}{status}{style:#} {msg}");
    }

    pub fn print_right(&self, status: &str, style: &Style, msg: impl Display) {
        let mut stream = AutoStream::new(stderr(), self.0);
        let _ = writeln!(stream, "{style}{status:>12}{style:#} {msg}");
    }

    pub fn update_right(&self, status: &str, style: &Style, msg: impl Display) {
        let mut stream = AutoStream::new(stderr(), self.0).lock();
        let _ = stream.execute(MoveToPreviousLine(1));
        let _ = writeln!(stream, "{style}{status:>12}{style:#} {msg}");
    }
}

static CONSOLE: OnceLock<Console> = OnceLock::new();

pub fn console() -> &'static Console {
    CONSOLE.get().expect("console should be initialized")
}

pub fn init(color: Color) {
    CONSOLE
        .set(Console::new(color))
        .expect("console should only be initialized once")
}

macro_rules! error {
    ($($arg:tt)+) => (
        crate::console::console().print(
            "error:",
            &crate::console::ERROR,
            format_args!($($arg)+),
        )
    );
}

macro_rules! info {
    ($status:expr, $($arg:tt)+) => (
        crate::console::console().print_right(
            $status,
            &crate::console::GOOD,
            format_args!($($arg)+),
        )
    );
}

macro_rules! fail {
    ($($arg:tt)+) => {
        crate::console::console().update_right("Failed", &crate::console::ERROR, "");
        error!($($arg)+);
    }
}

macro_rules! update {
    ($status:expr, $($arg:tt)+) => (
        crate::console::console().update_right(
            $status,
            &crate::console::GOOD,
            format_args!($($arg)+),
        )
    );
}

pub(crate) use error;
pub(crate) use fail;
pub(crate) use info;
pub(crate) use update;
