use std::{env::consts::DLL_EXTENSION, fs::create_dir_all, path::Path, process::ExitCode};

use clap::Parser;
use clap_cargo::style::CLAP_STYLING;
use git2::Repository;

mod cc;
mod console;
mod languages;
mod settings;

use console::{Color, error, fail, info, update};
use languages::{InstallInfo, get_install_info};
use settings::{Settings, get_settings};

#[derive(Debug, Parser)]
#[command(version, about, long_about = None)]
#[command(styles = CLAP_STYLING)]
struct Cli {
    /// when to use terminal colors
    #[arg(short, long, value_enum, default_value_t = Color::Auto)]
    color: Color,

    languages: Vec<String>,
}

fn main() -> ExitCode {
    let cli = Cli::parse();

    console::init(cli.color);

    match install_all(cli.languages) {
        Err(e) => {
            error!("{e}");
            ExitCode::FAILURE
        }
        Ok(code) => code,
    }
}

fn install_all(mut languages: Vec<String>) -> anyhow::Result<ExitCode> {
    let Settings {
        parser_dir,
        compiler,
    } = get_settings()?;

    let src_dir = tempfile::tempdir()?;

    create_dir_all(&parser_dir)?;

    // have to use a canonical path bc compiler runs elsewhere
    let parser_dir = parser_dir.canonicalize()?;

    let mut errors = 0;
    let mut installed = 0;

    languages.sort();
    languages.dedup();

    for language in &languages {
        if let Some(info) = get_install_info(language) {
            if let Err(e) = install(
                language,
                info,
                &parser_dir,
                src_dir.as_ref(),
                compiler.as_deref(),
            ) {
                fail!("{e}");
                if e.fatal() {
                    return Ok(ExitCode::FAILURE);
                } else {
                    errors += 1;
                }
            } else {
                installed += 1;
            }
        } else {
            error!("unknown language `{language}`");
        }
    }

    info!("Finished", "installed {installed} parsers");

    Ok(if errors > 0 {
        ExitCode::FAILURE
    } else {
        ExitCode::SUCCESS
    })
}

#[derive(Debug, thiserror::Error)]
enum InstallError {
    #[error("couldn't clone `{0}`")]
    Clone(&'static str),

    #[error(transparent)]
    Git(#[from] git2::Error),

    #[error(transparent)]
    Compile(#[from] cc::Error),
}

impl InstallError {
    fn fatal(&self) -> bool {
        match self {
            Self::Compile(e) => e.fatal(),
            _ => false,
        }
    }
}

fn install(
    language: &str,
    InstallInfo { url, files }: &InstallInfo,
    parser_dir: &Path,
    src_dir: &Path,
    compiler: Option<&str>,
) -> Result<(), InstallError> {
    info!("Installing", "{language}");

    let repo_path = src_dir.join(language);
    let repo = Repository::clone(url, &repo_path).map_err(|_| InstallError::Clone(url))?;

    let version = repo.head()?.peel_to_commit()?.id();

    let output_path = parser_dir.join(language).with_extension(DLL_EXTENSION);

    let mut build = cc::Build::new();
    build.compiler(compiler).dir(&repo_path).input_files(files);
    build.compile(&output_path)?;

    update!("Installed", "{language} {version:.7}");

    Ok(())
}
